#include <MaskedScalerCalibFilter.h>
#include <highgui.h>
#include <stdio.h> 
#include <stdlib.h>

// Dominik Schmidt (2008)
// schmidtd@comp.lancs.ac.uk
// http://eis.comp.lancs.ac.uk/~dominik/cms


const char *MaskedScalerCalibFilter::PARAM_SCALER_FILENAME = "scalerFilename";
const char *MaskedScalerCalibFilter::PARAM_RECTIFY_FILENAME = "rectifyFilename";
const char *MaskedScalerCalibFilter::PARAM_STEP_SIZE = "stepSize"; 
const char *MaskedScalerCalibFilter::TRACKBAR_LABEL_SCALER_LEVEL = "scaler"; 
const char *MaskedScalerCalibFilter::TRACKBAR_LABEL_RECTIFY_LEVEL = "rectify"; 
const CvScalar MaskedScalerCalibFilter::COLOR_ACTIVE_POINT = cvScalar(255, 255, 255);
const CvScalar MaskedScalerCalibFilter::COLOR_INACTIVE_POINT = cvScalar(128, 128, 128);
const int MaskedScalerCalibFilter::POINT_SIZE = 10;
const int MaskedScalerCalibFilter::DEFAULT_CALIB_LEVEL = 255;
const int MaskedScalerCalibFilter::DEFAULT_RECTIFY_LEVEL = 120;
const int MaskedScalerCalibFilter::DEFAULT_STEP_SIZE = 80;
const int MaskedScalerCalibFilter::MODE_CALIB = 0;
const int MaskedScalerCalibFilter::MODE_SCALER_MAP = 1;
const int MaskedScalerCalibFilter::MODE_RECTIFY_MAP = 2;
const int MaskedScalerCalibFilter::MODE_LIVE = 3;

// static functions and variables
int globalActivePointOffset = 0;
int globalCycleMode = 0;
int globalAutoCalib = 0;
// Left button: go to next calib point
// CTRL + left button: cycle through display modes
// Right button down: auto calib start
// Right button up: auto calin stop
// ALT + right button: go to previous calib point
void mouseHandler(int event, int x, int y, int flags, void* param) {
	switch(event){
		case CV_EVENT_RBUTTONDOWN:
			if(!(flags & CV_EVENT_FLAG_ALTKEY))
				globalAutoCalib = 1;
			break;
		case CV_EVENT_LBUTTONUP:			
			if(flags & CV_EVENT_FLAG_CTRLKEY)  {
				globalCycleMode = 1;
			}
			else {
				globalActivePointOffset++;
			}
			break;
		case CV_EVENT_RBUTTONUP:
			globalAutoCalib = 0;
			if(flags & CV_EVENT_FLAG_ALTKEY)
				globalActivePointOffset--;			
			break;
	}
}

MaskedScalerCalibFilter::MaskedScalerCalibFilter(char* s) : Filter(s) {
	initDone = 0;
	mode = MaskedScalerCalibFilter::MODE_CALIB;
	stepSize = DEFAULT_STEP_SIZE;
}

MaskedScalerCalibFilter::~MaskedScalerCalibFilter() {
	
}

void MaskedScalerCalibFilter::getParameters(ParameterMap& pMap) {
	pMap[std::string(PARAM_SCALER_FILENAME)] = toString(scalerFilename);
	pMap[std::string(PARAM_RECTIFY_FILENAME)] = toString(rectifyFilename);
	pMap[std::string(PARAM_STEP_SIZE)] = toString(stepSize);	
}

void MaskedScalerCalibFilter::setParameter(const char *name, const char *value) {
	if (strcmp(name, PARAM_SCALER_FILENAME) == 0) {
		scalerFilename = new char[255];
		sprintf(scalerFilename, "%s", value);					
	}
	else if (strcmp(name, PARAM_RECTIFY_FILENAME) == 0) {
		rectifyFilename = new char[255];
		sprintf(rectifyFilename, "%s", value);					
	}
	else if (strcmp(name, PARAM_STEP_SIZE) == 0) {
		stepSize = atoi(value);					
	}
}

void MaskedScalerCalibFilter::showOutput(bool value, int windowx, int windowy) {
	Filter::showOutput(value, windowx, windowy);

	if (value) {
		cvCreateTrackbar(TRACKBAR_LABEL_SCALER_LEVEL, name.c_str(), &scalerLevel, 255, NULL);
		cvCreateTrackbar(TRACKBAR_LABEL_RECTIFY_LEVEL, name.c_str(), &rectifyLevel, 255, NULL);
	}

	int mouseParam=5;
	cvSetMouseCallback(this->name.c_str(), mouseHandler, &mouseParam);
}

void MaskedScalerCalibFilter::kernel() {
	if(initDone == 0)
		init();

	if(!destination) {
        destination = cvCreateImage(cvSize(source->width,source->height), source->depth, 1);
        destination->origin = source->origin;
    }

	// check if user requested mode change
	cycleMode();
	switch(mode) {
		case MaskedScalerCalibFilter::MODE_CALIB:
			renderModeCalib();
			break;
		case MaskedScalerCalibFilter::MODE_SCALER_MAP:
			renderModeScalerMap();
			break;
		case MaskedScalerCalibFilter::MODE_RECTIFY_MAP:
			renderModeRectifyMap();
			break;
		case MaskedScalerCalibFilter::MODE_LIVE:
			renderModeLive();
			break;
	}
	
}

void MaskedScalerCalibFilter::cycleMode() {
	if(globalCycleMode == 1) {
		storeCalibValue();

		switch(mode) {
			case MaskedScalerCalibFilter::MODE_CALIB:					
				mode = MaskedScalerCalibFilter::MODE_SCALER_MAP;
				break;
			case MaskedScalerCalibFilter::MODE_SCALER_MAP:
				mode = MaskedScalerCalibFilter::MODE_RECTIFY_MAP;
				break;
			case MaskedScalerCalibFilter::MODE_RECTIFY_MAP:
				mode = MaskedScalerCalibFilter::MODE_LIVE;
				break;
			case MaskedScalerCalibFilter::MODE_LIVE:
				mode = MaskedScalerCalibFilter::MODE_CALIB;
				break;
		}
		globalCycleMode = 0;
	}
}

void MaskedScalerCalibFilter::renderModeCalib() {
	cyclePoint();

	if(globalAutoCalib == 1)
		autoCalibrate(source, activePoint, 20);

	applyScaler(source, destination, scalerLevel);
	applyRectify(destination, destination, rectifyLevel);
	

	drawCalibUI(destination);	
}

void MaskedScalerCalibFilter::autoCalibrate(IplImage* _src, int calibPoint, int calibAreaSize) {
	IplImage* src = cvCreateImage(cvSize(_src->width, _src->height), _src->depth, 1);

	int calibSteps = 0;
	int maxCalibSteps = 512;
	int calibrated = 0;
	int canceled = 0;
	int calibState = 0;
	int targetAreaSize = 40;
	int targetAreaTolerance = 4;

	int initialScaler = DEFAULT_CALIB_LEVEL;
	int initialRectify = 254;
	
	int calibratedScaler = initialScaler;
	int calibratedRectify = initialRectify;

	int area;
	do {
		// Restore orginal source image
		cvResetImageROI(src);
		unsigned long len = src->imageSize;
		memcpy(src->imageData, _src->imageData, len);

		// Set ROI
		CvPoint calibPointCoord = getCalibPointCoords(calibPoint);
		CvRect roi = cvRect(calibPointCoord.x - calibAreaSize / 2, 
			calibPointCoord.y - calibAreaSize / 2, 
			calibAreaSize, calibAreaSize);
		cvSetImageROI(src, roi);

		// Apply filters with current values
		applyScaler(src, src, calibratedScaler);
		applyRectify(src, src, calibratedRectify);

		area = detectBlobs(src);
		switch(calibState) {
			// Initial
			case 0:
				// Nothing detected
				if(area == 0) {
					if(calibratedRectify > 0)
						calibratedRectify--;
					else
						canceled = 1;
				}
				// Too many blobs detected
				else if(area == -1) {
					if(calibratedScaler > 0)
						calibratedScaler--;
					else
						canceled = 1;
				}
				// One blob deteced
				else {
					calibState = 1;
				}
				break;
			// One blob
			case 1:
				if(area == 0 || area == -1)
					canceled = 1;
				else {
					if(fabs((float) (area - targetAreaSize)) <= targetAreaTolerance) 
						calibrated = 1;
					else {
						// Detected area is too small
						if(area < targetAreaSize) {
							if(calibratedRectify > 0)
								calibratedRectify--;
							else
								canceled = 1;
						}
						// Detected area is too large
						else if(area > targetAreaSize) {
							if(calibratedScaler > 0)
								calibratedScaler--;
							else
								canceled = 1;
							}
					}
				}
				break;
		}		

		calibSteps++;
	} while(calibSteps < maxCalibSteps && calibrated == 0 && canceled == 0);

	printf("autocalib: calibrated=%i, canceled=%i, calibratedScaler=%i, calibratedRectify=%i, area=%i\n", calibrated, canceled, calibratedScaler, calibratedRectify, area);

	if(calibrated == 1) {
		scalerLevel = calibratedScaler;
		rectifyLevel = calibratedRectify;
		scalerCalibValues[calibPoint] = scalerLevel;
		rectifyCalibValues[calibPoint] = rectifyLevel;
		updateTrackbar();
	}

	cvReleaseImage(&src);
}

int MaskedScalerCalibFilter::detectBlobs(IplImage* img) {
	int area;


	CvMemStorage* storage = cvCreateMemStorage(0);
	CvSeq* cont = 0; 
	CvSeq *result;				
	
	cvFindContours(img, storage, &cont, sizeof(CvContour), CV_RETR_TREE, CV_CHAIN_APPROX_SIMPLE);

	// check that there is only 1 blob
	CvSeq* firstCont = cont;
	int numberOfContours = 0;
	for( ; cont != 0; cont = cont->h_next )	
		numberOfContours++;


	if(numberOfContours == 1) {
		// get area of blob
		cont = firstCont;								
		result = cvApproxPoly( cont, sizeof(CvContour), storage,
		CV_POLY_APPROX_DP, cvContourPerimeter(cont)*0.02, 0 );
		area = (int) fabs(cvContourArea(result,CV_WHOLE_SEQ));	
	} else if(numberOfContours == 0)
		area = 0;
	else
		area = -1;
			 		
	cvReleaseMemStorage(&storage);
	return area;
}

void MaskedScalerCalibFilter::renderModeScalerMap() {
	unsigned long len = destination->imageSize;
	memcpy(destination->imageData, scalerMapImagePreview->imageData, len);
}

void MaskedScalerCalibFilter::renderModeRectifyMap() {
	unsigned long len = destination->imageSize;
	memcpy(destination->imageData, rectifyMapImagePreview->imageData, len);
}

void MaskedScalerCalibFilter::renderModeLive() {
	applyMaskedScaler(source, destination, scalerMapImagePreview);	
	applyMaskedRectify(destination, destination, rectifyMapImagePreview);
}


void MaskedScalerCalibFilter::init() {
	imgSize = cvSize(source->width, source->height);
	steps = cvPoint(imgSize.width / stepSize,
		imgSize.height / stepSize);

	// check if another point can be added on the borders
	if(imgSize.width % stepSize == 0)
		steps.x++;
	if(imgSize.height % stepSize == 0)
		steps.y++;

	nCalibPoints = steps.x * steps.y;
	activePoint = 0;

	scalerCalibValues = (int*) malloc(nCalibPoints * sizeof(int));
	for(int i = 0; i < nCalibPoints; i++)
		scalerCalibValues[i] = MaskedScalerCalibFilter::DEFAULT_CALIB_LEVEL;

	rectifyCalibValues = (int*) malloc(nCalibPoints * sizeof(int));
	for(int i = 0; i < nCalibPoints; i++)
		rectifyCalibValues[i] = MaskedScalerCalibFilter::DEFAULT_RECTIFY_LEVEL;
	
	scalerLevel = scalerCalibValues[activePoint];	
	rectifyLevel = rectifyCalibValues[activePoint];	
	updateTrackbar();
	
	loadResults(scalerCalibValues, scalerFilename);
	loadResults(rectifyCalibValues, rectifyFilename);
	scalerMapImagePreview = interpolateMap(scalerCalibValues);
	rectifyMapImagePreview = interpolateMap(rectifyCalibValues);

	initDone = 1;
}

CvPoint MaskedScalerCalibFilter::getCalibPointCoords(int i) {	
	assert(i >= 0 && i < nCalibPoints);

	int yIndex = i / (steps.x);
	int xIndex = i - yIndex * (steps.x);
	
	return cvPoint(xIndex * stepSize,
		yIndex * stepSize);
}

void MaskedScalerCalibFilter::cyclePoint() {
	if(globalActivePointOffset != 0) {
		storeCalibValue();

		activePoint = (activePoint + globalActivePointOffset) % nCalibPoints;
		globalActivePointOffset = 0;
		if(activePoint < 0)
			activePoint = nCalibPoints - 1;

		updateTrackbar();
	}
}


void MaskedScalerCalibFilter::storeCalibValue() {
	scalerCalibValues[activePoint] = scalerLevel;
	rectifyCalibValues[activePoint] = rectifyLevel;

	cvReleaseImage(&scalerMapImagePreview);
	cvReleaseImage(&rectifyMapImagePreview);
	scalerMapImagePreview = interpolateMap(scalerCalibValues);
	rectifyMapImagePreview = interpolateMap(rectifyCalibValues);
	saveResults(scalerCalibValues, scalerMapImagePreview, scalerFilename);
	saveResults(rectifyCalibValues, rectifyMapImagePreview, rectifyFilename);
}

void MaskedScalerCalibFilter::updateTrackbar() {
	cvSetTrackbarPos(TRACKBAR_LABEL_SCALER_LEVEL, this->name.c_str(), scalerCalibValues[activePoint]);
	cvSetTrackbarPos(TRACKBAR_LABEL_RECTIFY_LEVEL, this->name.c_str(), rectifyCalibValues[activePoint]);
}

void MaskedScalerCalibFilter::applyRectify(IplImage* src, IplImage* dst, int level) {
	// copied from RectifyFilter.cpp
	cvThreshold(src, dst, level, 255, CV_THRESH_TOZERO);
}

void MaskedScalerCalibFilter::applyScaler(IplImage* src, IplImage* dst, int level) {
	// copied from ScalerFilter.cpp
	cvMul(src, src, dst, (float) level / 128.0f);
}

void MaskedScalerCalibFilter::applyMaskedScaler(IplImage* src, IplImage* dst, IplImage* mask) {
	// copied from MaskedScalerFilter.cpp
	int value;
	for(int y = 0; y < source->height; y++) {		
		uchar* ptrDes = (uchar*) (
			dst->imageData + y * dst->widthStep
		);
		uchar* ptrSrc = (uchar*) (
			src->imageData + y * src->widthStep
		);
		uchar* ptrMas = (uchar*) (
			mask->imageData + y * mask->widthStep
		);
		for(int x = 0; x < source->width; x++) {	
			value = (int) (ptrSrc[x] * ptrSrc[x]  * ((float) ptrMas[x] / 128.0f));
			if(value > 255) value = 255;

			ptrDes[x] = value;			
		}
	}
}

void MaskedScalerCalibFilter::applyMaskedRectify(IplImage* src, IplImage* dst, IplImage* mask) {
	for(int y = 0; y < source->height; y++) {		
		uchar* ptrDes = (uchar*) (
			dst->imageData + y * dst->widthStep
		);
		uchar* ptrSrc = (uchar*) (
			src->imageData + y * src->widthStep
		);
		uchar* ptrMas = (uchar*) (
			mask->imageData + y * mask->widthStep
		);		
		for(int x = 0; x < source->width; x++) {	
			if(ptrSrc[x] <= ptrMas[x])
				ptrDes[x] = 0;
			else {
				ptrDes[x] = ptrSrc[x];				
			}
		}
	}
}


void MaskedScalerCalibFilter::drawCalibPoint(IplImage* img, CvPoint p, int active) {	
	if(active)
		cvCircle(img, p, MaskedScalerCalibFilter::POINT_SIZE, 
			MaskedScalerCalibFilter::COLOR_ACTIVE_POINT, 2); 
	else
		cvCircle(img, p, MaskedScalerCalibFilter::POINT_SIZE, 
			MaskedScalerCalibFilter::COLOR_INACTIVE_POINT, 2); 
}

void MaskedScalerCalibFilter::drawCalibUI(IplImage* img) {
	for(int i = 0; i < nCalibPoints; i++) {
		CvPoint p = getCalibPointCoords(i);
		drawCalibPoint(img, p, activePoint == i);
	}
}

void MaskedScalerCalibFilter::loadResults(int* calibValues, char* _fileName) {
	char* fileName = new char[255];
	sprintf(fileName, "%s.dat", _fileName);

	FILE *file;
	file = fopen(fileName, "r");

	char in [255];
	if(file != NULL) {
		fgets(in , 255 , file);
		int fileCalibPoints = atoi(in);

		if(fileCalibPoints == nCalibPoints) {
			for(int i = 0; i < nCalibPoints; i++) {
				fgets(in, 255, file);
				calibValues[i] = atoi(in);
			}			
		} else {
			printf("wrong number of calibration points in file %s\n", fileName);
		}
		fclose(file);
	}
}

void MaskedScalerCalibFilter::saveResults(int* calibValues, IplImage* image, char* _fileName) {	
	char* fileName = new char[255];
	sprintf(fileName, "%s.png", _fileName);
	cvSaveImage(fileName, image);

	sprintf(fileName, "%s.dat", _fileName);
	FILE *file;
	file = fopen(fileName, "w");
	fprintf(file, "%i\n", nCalibPoints);
	for(int i = 0; i < nCalibPoints; i++) {
		fprintf(file, "%i\n", calibValues[i]);
	}
	fclose(file); 	
}

// #############################################################
// MAP INTERPOLATION
// #############################################################

// assuming point k within a triangel ABC with A(0/0),
// B(s/0), and C(0/s)
double MaskedScalerCalibFilter::getInterpolatedValueUpperTriangle(CvPoint k, CvScalar values, int s) {
	// just interpolating between A and B
	if(k.y == 0)
		return ((double) (s - k.x) / (double) s) * values.val[0] + ((double) k.x / (double) s) * values.val[1];

	if(s - k.x != 0) {
		double angle = atan((double) k.y / (double) (s - k.x));
		double ty = tan(angle) * (double) s;

		double valueTemp = ((s - ty) / s) * values.val[0] + (ty / s) * values.val[2];		

		double l1 = ty / sin(angle);
		double l2 = k.y / sin(angle);

		double value =  (l2 / l1) * valueTemp + ((l1 - l2) / l1) * values.val[1];

		return value;
	} else {		
		return values.val[0];
	}
}

// assuming point k within a triangel ABC with A(s/0),
// B(s/s), and C(0/s)
double MaskedScalerCalibFilter::getInterpolatedValueLowerTriangle(CvPoint k, CvScalar values, int s) {
	// just interpolating between A and B
	if(k.x == 0)
		return ((double) (s - k.y) / (double) s) * values.val[0] + ((double) k.y / (double) s) * values.val[1];

	if(s - k.y != 0) {
		double angle = atan((double) (s - k.y) / (double) k.x);
		double ty = tan(angle) * (double) s;

		double valueTemp = ((s - ty) / s) * values.val[1] + (ty / s) * values.val[0];

		double l1 = ty / sin(angle);
		double l2 = (s - k.y) / sin(angle);

		double value =  (l2 / l1) * valueTemp + ((l1 - l2) / l1) * values.val[2];

		return value;
	} else {		
		return values.val[1];
	}
}


// interpolates area of rectangle specified by the four points
void MaskedScalerCalibFilter::interpolateRectangle(IplImage* dest, int* calibValues, int p1Index, int p2Index, int p3Index, int p4Index) {
	CvPoint p1 = getCalibPointCoords(p1Index);
	CvPoint p2 = getCalibPointCoords(p2Index);
	CvPoint p3 = getCalibPointCoords(p3Index);
	CvPoint p4 = getCalibPointCoords(p4Index);

	int destImgStep = dest->widthStep / sizeof(uchar);
	uchar* data = (uchar*) dest->imageData;
	
	for(int y = 0; y < stepSize; y++) {
		int kx = stepSize - y;
		double value;
		for(int x = 0; x <= kx; x++) {			
			value = getInterpolatedValueUpperTriangle(cvPoint(x, y), 
				cvScalar(calibValues[p1Index], calibValues[p2Index], calibValues[p4Index]), stepSize);
			data[(p1.y + y) * destImgStep + (p1.x + x)] = (uchar) value;
		}
		for(int x = kx; x < stepSize; x++) {			
			value = getInterpolatedValueLowerTriangle(cvPoint(x, y), 
				cvScalar(calibValues[p2Index], calibValues[p3Index], calibValues[p4Index]), stepSize);
			data[(p1.y + y ) * destImgStep + (p1.x + x)] = (uchar) value;
		}				
	}

}

// interpolates over all points and creates output image map
IplImage* MaskedScalerCalibFilter::interpolateMap(int* calibValues) {
	IplImage* dest = cvCreateImage(imgSize, IPL_DEPTH_8U, 1);

	int p1Index, p2Index, p3Index, p4Index;
	for(int y = 0; y < steps.y - 1; y++) {
		for(int x = 0; x < steps.x - 1; x++) {
			p1Index = y * steps.x + x;
			p2Index = p1Index + 1;
			p4Index = (y + 1) * steps.x + x;
			p3Index = p4Index + 1;
			
			interpolateRectangle(dest, calibValues, p1Index, p2Index, p3Index, p4Index);			
		}
	}

	return dest;
}