#ifndef __TOUCHSCREEN_MASKED_FILTER_CALIB_SCALER__

#define __TOUCHSCREEN_MASKED_FILTER_CALIB_ SCALER__
#include <TouchlibFilter.h>

// Dominik Schmidt (2008)
// schmidtd@comp.lancs.ac.uk
// http://eis.comp.lancs.ac.uk/~dominik/cms

class TOUCHLIB_FILTER_EXPORT MaskedScalerCalibFilter : public Filter {
public:
	MaskedScalerCalibFilter();
	MaskedScalerCalibFilter(char* s);
	void kernel();
    virtual ~MaskedScalerCalibFilter();
	virtual void getParameters(ParameterMap& pMap);
	virtual void setParameter(const char *name, const char *value);
	virtual void showOutput(bool value, int windowx, int windowy);
private:	
	// config XML parameter for scaler map filename
	static const char *PARAM_SCALER_FILENAME;
	// config XML parameter for rectify map filename
	static const char *PARAM_RECTIFY_FILENAME;
	// config XML parameter for distance between  calibration points
	static const char *PARAM_STEP_SIZE;
	// label of trackbar for scaler level adjustment
	static const char *TRACKBAR_LABEL_SCALER_LEVEL;	
	// label of trackbar for rectify level adjustment
	static const char *TRACKBAR_LABEL_RECTIFY_LEVEL;
	// calibration points get initialized with this value
	static const int DEFAULT_CALIB_LEVEL;
	// calibration points get initialized with this value
	static const int DEFAULT_RECTIFY_LEVEL;
	// default step size between calibration points
	static const int DEFAULT_STEP_SIZE;


	// calibration mode
	static const int MODE_CALIB;
	// showing scaler map image
	static const int MODE_SCALER_MAP;
	// showing rectify map image
	static const int MODE_RECTIFY_MAP;
	// live preview
	static const int MODE_LIVE;

	// colors for calibration UI
	static const CvScalar COLOR_ACTIVE_POINT;
	static const CvScalar COLOR_INACTIVE_POINT;
	// size of calibration points
	static const int POINT_SIZE;

	// currently selected rectify level
	int rectifyLevel;
	// currently selected scaler level
	int scalerLevel;
	// filename used to load and save scaler results (and image map)
	char* scalerFilename;
	// filename used to load and save rectify results (and image map)
	char* rectifyFilename;

	// calibration, map preview, or live preview
	int mode;

	// set to 1 after init
	int initDone;

	// distance between calibration points (same for x and y)
	int stepSize;
	// size of input image
	CvSize imgSize;	
	// number of steps in x and y direction
	CvPoint steps;
	// number of calibration points
	int nCalibPoints;	
	// scaler values (result of calibration)
	int* scalerCalibValues;
	// rectify values (result of calibration)
	int* rectifyCalibValues;
	// index of active calibration point
	int activePoint;

	// used to show scaler map preview
	IplImage* scalerMapImagePreview;
	// used to show rectify map preview
	IplImage* rectifyMapImagePreview;

	// setting up handler etc.
	void init();

	// retrieves x and y coordinates for point with index i
	CvPoint getCalibPointCoords(int i);
	
	// reads trackbar and stores value for active point in array
	void storeCalibValue();
	// sets trackbar value to value of active point
	void updateTrackbar();

	// applies rectify filter
	void applyRectify(IplImage* src, IplImage* dst, int level);
	// applies scaler filter
	void applyScaler(IplImage* src, IplImage* dst, int level);
	// applies masked scaler filter
	void applyMaskedScaler(IplImage* src, IplImage* dst, IplImage* mask);
	// applies masked rectify filter
	void applyMaskedRectify(IplImage* src, IplImage* dst, IplImage* mask);

	// returns -1 if more than one blob was detected, 0 if there is no blob, and the blob's area otherwise
	int detectBlobs(IplImage* img);
	void autoCalibrate(IplImage* _src, int calibPoint, int calibAreaSize);

	// cycle through modes
	void cycleMode();
	// cycle through calibration points
	void cyclePoint();

	// called from kernel, rendering different outputs
	void renderModeCalib();
	void renderModeScalerMap();
	void renderModeRectifyMap();
	void renderModeLive();


	// visualizes calibration point at given coordinates, i.e. draw
	// a visual representation (depending on acitve state) onto 
	// the specified image
	void drawCalibPoint(IplImage* img, CvPoint p, int active);

	// visualizes calibration UI consisting of calibration points
	// and text output
	void drawCalibUI(IplImage* img);

	// saves results to file
	void saveResults(int* calibValues, IplImage* image, char* _fileName);
	// loads calibration from file
	void loadResults(int* calibValues, char* _fileName);
	
	// assuming point k within a triangel ABC with A(0/0),
	// B(s/0), and C(0/s)
	double getInterpolatedValueUpperTriangle(CvPoint k, CvScalar values, int s);

	// assuming point k within a triangel ABC with A(s/0),
	// B(s/s), and C(0/s)
	double getInterpolatedValueLowerTriangle(CvPoint k, CvScalar values, int s);

	// interpolates area of rectangle specified by the four points
	void interpolateRectangle(IplImage* dest, int* calibValues, int p1Index, int p2Index, int p3Index, int p4Index);

	// interpolates over all points and creates output image map
	IplImage* interpolateMap(int* calibValues);

};

#endif